# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE-POS -> Catapult Inventory Workbook
"""

import re
import datetime
import decimal
import logging

from sqlalchemy.exc import ProgrammingError
from sqlalchemy import orm
from sqlalchemy.orm.exc import NoResultFound

from corepos import enum as corepos_enum
from corepos.db.office_op import model as corepos
from corepos.db.util import table_exists

from rattail.gpc import GPC
from rattail.core import get_uuid
from rattail.util import OrderedDict
from rattail.importing.handlers import ToFileHandler
from rattail_corepos.corepos.importing.db.corepos import FromCoreHandler, FromCore
from rattail_onager.catapult.importing import inventory as catapult_importing


log = logging.getLogger(__name__)


class FromCoreToCatapult(FromCoreHandler, ToFileHandler):
    """
    Handler for CORE -> Catapult (Inventory Workbook)
    """
    host_title = "CORE-POS"
    local_title = "Catapult (Inventory Workbook)"
    direction = 'export'

    def get_importers(self):
        importers = OrderedDict()
        importers['InventoryItem'] = InventoryItemImporter
        return importers


class InventoryItemImporter(FromCore, catapult_importing.model.InventoryItemImporter):
    """
    Inventory Item data importer.
    """
    host_model_class = corepos.Product
    # note that we use a "dummy" uuid key here, so logic will consider each row
    # to be unique, even when duplicate item_id's are present
    key = 'uuid'
    supported_fields = [
        'uuid',
        'item_id',
        'dept_id',
        'dept_name',
        'receipt_alias',
        'brand',
        'item_name',
        'size',
        # 'sugg_retail',
        'last_cost',
        'price_divider',
        'base_price',
        # 'disc_mult',
        'ideal_margin',
        'bottle_deposit',
        # 'pos_menu_group',
        'scale_label',
        'sold_by_ea_or_lb',
        'quantity_required',
        'weight_profile',
        'tax_1',
        'tax_2',
        'spec_tend_1',
        'spec_tend_2',
        'age_required',
        'location',
        # 'family_line',
        'alt_id',
        'alt_receipt_alias',
        'alt_pkg_qty',
        'alt_pkg_price',
        'auto_discount',
        'supplier_unit_id',
        'supplier_id',
        'unit',
        'num_pkgs',
        # 'cs_pk_multiplier',
        # 'dsd',
        'pf1',
        # 'pf2',
        # 'pf3',
        # 'pf4',
        # 'pf5',
        # 'pf6',
        # 'pf7',
        # 'pf8',
        'memo',
        'scale_shelf_life',
        'scale_shelf_life_type',
        'scale_ingredient_text',
    ]

    # we want to add a "duplicate" column at the end
    include_duplicate_column = True

    # we want to add an "alternate for" column at the end
    include_alt_for_column = True

    type2_upc_pattern = re.compile(r'^2(\d{5})00000\d')

    def setup(self):
        super(InventoryItemImporter, self).setup()

        # this is used for sorting, when a value has no date
        self.old_datetime = datetime.datetime(1900, 1, 1)

        self.exclude_invalid_upc = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.exclude_invalid_upc',
            default=False)

        self.warn_invalid_upc = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_invalid_upc',
            default=True)

        self.ignored_upcs = self.config.getlist(
            'corepos', 'exporting.catapult_inventory.ignored_upcs')

        self.exclude_missing_department = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.exclude_missing_department',
            default=False)

        self.warn_missing_department = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_missing_department',
            default=True)

        self.warn_empty_subdepartment = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_empty_subdepartment',
            default=True)

        self.warn_truncated_receipt_alias = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_truncated_receipt_alias',
            default=True)

        self.warn_size_null_byte = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_size_null_byte',
            default=True)

        self.warn_unknown_deposit = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_unknown_deposit',
            default=True)

        self.warn_scale_label_non_plu = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_scale_label_non_plu',
            default=True)

        self.warn_scale_label_short_plu = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_scale_label_short_plu',
            default=True)

        self.warn_weight_profile_non_plu = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_weight_profile_non_plu',
            default=True)

        self.warn_multiple_vendor_items = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_multiple_vendor_items',
            default=True)

        self.warn_no_valid_vendor_items = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_no_valid_vendor_items',
            default=True)

        self.warn_truncated_memo = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_truncated_memo',
            default=True)

        self.warn_scale_ingredients_newline = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_scale_ingredients_newline',
            default=True)

        self.floor_sections_exist = table_exists(self.host_session,
                                                 corepos.FloorSection)
        self.tax_components_exist = table_exists(self.host_session,
                                                 corepos.TaxRateComponent)

        self.tax_rate_ids_1 = self.config.getlist(
            'corepos', 'exporting.catapult_inventory.tax_rate_ids_1', default=[])
        self.tax_rate_ids_1 = [int(id) for id in self.tax_rate_ids_1]
        self.tax_rate_ids_2 = self.config.getlist(
            'corepos', 'exporting.catapult_inventory.tax_rate_ids_2', default=[])
        self.tax_rate_ids_2 = [int(id) for id in self.tax_rate_ids_2]

        # TODO: should add component id levels too?
        # tax_component_ids_1 = (1,)
        # tax_component_ids_2 = (2,)

        self.cache_bottle_deposits()
        self.cache_like_codes()

    def cache_bottle_deposits(self):
        self.deposits = {}
        deposits = self.host_session.query(corepos.Product.deposit.distinct())\
                                    .all()

        def cache(deposit, i):
            assert isinstance(deposit, tuple)
            assert len(deposit) == 1
            deposit = deposit[0]
            if deposit:
                deposit = int(deposit)
                upc = "{:013d}".format(deposit)
                try:
                    product = self.host_session.query(corepos.Product)\
                                               .filter(corepos.Product.upc == upc)\
                                               .one()
                except NoResultFound:
                    pass        # we will log warnings per-item later
                else:
                    self.deposits[deposit] = product

        self.progress_loop(cache, deposits,
                           message="Caching product deposits data")

    def cache_like_codes(self):
        self.like_codes = {}
        mappings = self.host_session.query(corepos.ProductLikeCode)\
                                    .order_by(corepos.ProductLikeCode.like_code_id,
                                              corepos.ProductLikeCode.upc)\
                                    .all()

        def cache(mapping, i):
            self.like_codes.setdefault(mapping.like_code_id, []).append(mapping)

        self.progress_loop(cache, mappings,
                           message="Caching like codes data")

    def query(self):
        query = self.host_session.query(corepos.Product)\
                                 .order_by(corepos.Product.upc)\
                                 .options(orm.joinedload(corepos.Product.department))\
                                 .options(orm.joinedload(corepos.Product.subdepartment))\
                                 .options(orm.joinedload(corepos.Product.vendor_items)\
                                          .joinedload(corepos.VendorItem.vendor))\
                                 .options(orm.joinedload(corepos.Product.default_vendor))\
                                 .options(orm.joinedload(corepos.Product.scale_item))\
                                 .options(orm.joinedload(corepos.Product.user_info))\
                                 .options(orm.joinedload(corepos.Product.tax_rate))\
                                 .options(orm.joinedload(corepos.Product._like_code))
        if self.floor_sections_exist:
            query = query.options(orm.joinedload(corepos.Product.physical_location)\
                                  .joinedload(corepos.ProductPhysicalLocation.floor_section))
        return query

    def normalize_host_data(self, host_objects=None):
        normalized = super(InventoryItemImporter, self).normalize_host_data(host_objects=host_objects)

        # re-sort the results by item_id, since e.g. original UPC from CORE may
        # have been replaced with a PLU.  also put non-numeric first, to bring
        # them to user's attention
        numeric = []
        non_numeric = []
        for row in normalized:
            if row['item_id'] and row['item_id'].isdigit():
                numeric.append(row)
            else:
                non_numeric.append(row)
        numeric.sort(key=lambda row: int(row['item_id']))
        non_numeric.sort(key=lambda row: row['item_id'])
        normalized = non_numeric + numeric

        # now we must check for duplicate item ids, and mark rows accordingly.
        # but we *do* want to include/preserve all rows, hence we mark them
        # instead of pruning some out. first step is to group all by item_id
        items = {}

        def collect(row, i):
            items.setdefault(row['item_id'], []).append(row)

        self.progress_loop(collect, normalized,
                           message="Grouping rows by Item ID")

        # now we go through our groupings and for any item_id with more than 1
        # row, we'll mark each row as having a duplicate item_id.  note that
        # this modifies such a row "in-place" for our overall return value
        def inspect(rows, i):
            if len(rows) > 1:
                for row in rows:
                    row['__duplicate__'] = True

        self.progress_loop(inspect, list(items.values()),
                           message="Marking any duplicate Item IDs")

        # finally, we must inspect the like codes and figure out which
        # product(s) should potentially be considered "alternate for" another.
        # first step here will be to create mapping of item_id values for each
        # CORE product in our result set
        item_ids = {}

        def mapp(row, i):
            product = row['__product__']
            item_ids[product.upc] = row['item_id']

        self.progress_loop(mapp, normalized,
                           message="Mapping item_id for CORE products")

        # next step here is to check each product and mark "alt for" as needed
        def inspect(row, i):
            product = row['__product__']
            if product.like_code:
                others = self.like_codes.get(product.like_code.id)
                if others:
                    first = others[0]
                    if first.upc != product.upc:
                        row['__alternate_for__'] = item_ids[first.upc]

        self.progress_loop(inspect, normalized,
                           message="Marking any \"alternate for\" items")

        return normalized

    def normalize_host_object(self, product):
        item_id = product.upc

        if self.ignored_upcs and item_id in self.ignored_upcs:
            log.debug("ignoring UPC %s for product: %s", product.upc, product)
            return

        if not item_id:
            logger = log.warning if self.warn_invalid_upc else log.debug
            logger("product id %s has no upc: %s", product.id, product)
            if self.exclude_invalid_upc:
                return

        if not item_id.isdigit():
            logger = log.warning if self.warn_invalid_upc else log.debug
            logger("product %s has non-numeric upc: %s",
                   product.upc, product)
            if self.exclude_invalid_upc:
                return

        # convert item_id either to a PLU, or formatted UPC
        is_plu = False
        if item_id.isdigit():   # can only convert if it's numeric!
            if len(str(int(item_id))) < 6:
                is_plu = True
                item_id = str(int(item_id))
            else: # must add check digit, and re-format
                upc = GPC(item_id, calc_check_digit='upc')
                item_id = str(upc)
                assert len(item_id) == 14
                # drop leading zero(s)
                if item_id[1] == '0': # UPC-A
                    item_id = item_id[2:]
                    assert len(item_id) == 12
                else: # EAN13
                    item_id = item_id[1:]
                    assert len(item_id) == 13

        # figure out the "scale label" data, which may also affect item_id
        scale_item = product.scale_item
        scale_label = None
        if scale_item:
            scale_label = 'Y'
            if item_id.isdigit():
                if len(item_id) < 5:
                    logger = log.warning if self.warn_scale_label_short_plu else log.debug
                    logger("product %s has scale label, but PLU is less than 5 digits (%s): %s",
                           product.upc, item_id, product)
                elif len(item_id) > 5:
                    match = self.type2_upc_pattern.match(item_id)
                    if match:
                        # convert type-2 UPC to PLU
                        is_plu = True
                        item_id = str(int(match.group(1)))
                        log.debug("converted type-2 UPC %s to PLU %s for: %s",
                                  product.upc, item_id, product)
                    else:
                        logger = log.warning if self.warn_scale_label_non_plu else log.debug
                        logger("product %s has scale label, but non-PLU item_id: %s",
                               product.upc, product)

        department = product.department
        if not department:
            logger = log.warning if self.warn_missing_department else log.debug
            logger("product %s has no department: %s", product.upc, product)
            if self.exclude_missing_department:
                return

        # size may come from one of two fields, or combination thereof
        pack_size = (product.size or '').strip()
        uom = (product.unit_of_measure or '').strip()
        numeric_pack = False
        if pack_size:
            try:
                decimal.Decimal(pack_size)
            except decimal.InvalidOperation:
                pass
            else:
                numeric_pack = True
        if numeric_pack:
            size = "{} {}".format(pack_size, uom).strip()
        else:
            size = pack_size or uom or None
        # TODO: this logic may actually be client-specific?  i just happened to
        # find some null chars in a client DB and needed to avoid them, b/c the
        # openpyxl lib said IllegalCharacterError
        if size is not None and '\x00' in size:
            logger = log.warning if self.warn_size_null_byte else log.debug
            logger("product %s has null byte in size field: %s",
                   product.upc, product)
            size = size.replace('\x00', '')

        price_divider = None
        if (product.quantity and product.group_price and
            product.price_method == corepos_enum.PRODUCT_PRICE_METHOD_ALWAYS):
            diff = (product.quantity * product.normal_price) - product.group_price
            if abs(round(diff, 2)) > .01:
                log.warning("product %s has multi-price with $%0.2f diff: %s",
                            product.upc, diff, product)
            price_divider = product.quantity

        bottle_deposit = None
        if product.deposit:
            deposit = int(product.deposit)
            if deposit in self.deposits:
                bottle_deposit = self.deposits[deposit].normal_price
            else:
                logger = log.warning if self.warn_unknown_deposit else log.debug
                logger("product %s has unknown deposit %s which will be ignored: %s",
                       product.upc, deposit, product)

        sold_by_ea_or_lb = None
        if is_plu:
            sold_by_ea_or_lb = 'LB' if product.scale else 'EA'

        weight_profile = None
        if product.scale or scale_item:
            if not is_plu:
                logger = log.warning if self.warn_weight_profile_non_plu else log.debug
                logger("product %s has weight profile, but non-PLU item_id %s: %s",
                       product.upc, item_id, product)
            weight_profile = 'LBNT'

        # calculate tax rates according to configured "mappings"
        tax_1 = 0
        tax_2 = 0
        if product.tax_rate:

            # TODO: need to finish the logic to handle tax components
            if self.tax_components_exist and product.tax_rate.components:
                # for component in product.tax_rate.components:
                #     if tax_component_ids_1 and component.id in tax_component_ids_1:
                #         tax_1 += component.rate
                #     if tax_component_ids_2 and component.id in tax_component_ids_2:
                #         tax_2 += component.rate
                raise NotImplementedError

            else: # no components
                rate = product.tax_rate
                if self.tax_rate_ids_1 and rate.id in self.tax_rate_ids_1:
                    tax_1 += rate.rate
                if self.tax_rate_ids_2 and rate.id in self.tax_rate_ids_2:
                    tax_2 += rate.rate
                if not (self.tax_rate_ids_1 or self.tax_rate_ids_2) and rate.rate:
                    log.warning("product %s has unknown tax rate %s (%s) which will "
                                "be considered as tax 1: %s",
                                product.upc, rate.rate, rate.description, product)
                    tax_1 += rate.rate

        location = None
        if self.floor_sections_exist and product.physical_location and product.physical_location.floor_section:
            location = product.physical_location.floor_section.name
            if len(location) > 30:
                log.warning("product %s has location length %s; will truncate: %s",
                            product.upc, len(location), location)
                location = location[:30]

        # no alt item (or auto discount) by default
        alt_id = None
        alt_receipt_alias = None
        alt_pkg_qty = None
        alt_pkg_price = None
        auto_discount = None

        # make an alt item, when main item has pack pricing (e.g. Zevia sodas)
        # note that in this case the main item_id and alt_id are the same
        if (product.quantity and product.group_price and
            product.price_method == corepos_enum.PRODUCT_PRICE_METHOD_FULL_SETS):
            alt_id = item_id
            suffix = "{}-PK".format(product.quantity)
            alt_receipt_alias = "{} {}".format(product.description, suffix)
            if len(alt_receipt_alias) > 32:
                logger = log.warning if self.warn_truncated_receipt_alias else log.debug
                logger("alt receipt alias for %s is %s chars; must truncate: %s",
                       alt_id, len(alt_receipt_alias), alt_receipt_alias)
                overage = len(alt_receipt_alias) - 32
                alt_receipt_alias = "{} {}".format(
                    product.description[:-overage], suffix)
                assert len(alt_receipt_alias) == 32
            alt_pkg_qty = product.quantity
            alt_pkg_price = product.group_price

            # we also must declare an "auto discount" to get pack price
            auto_discount = "{} @ ${:0.2f}".format(alt_pkg_qty, alt_pkg_price)

        # no supplier info by default
        supplier_unit_id = None
        supplier_id = None
        supplier_unit = None
        supplier_num_pkgs = None

        # maybe add supplier info, for "default" `vendorItems` record.  we'll
        # have to get a little creative to figure out which is the default
        vendor_items = []

        # first we try to narrow down according to product's default vendor
        if product.default_vendor:
            vendor_items = [item for item in product.vendor_items
                            if item.vendor is product.default_vendor]

        # but if that didn't work, just use any "valid" vendorItems
        if not vendor_items:
            # valid in this context means, not missing vendor
            vendor_items = [item for item in product.vendor_items
                            if item.vendor]
            if not vendor_items and product.vendor_items:
                logger = log.warning if self.warn_no_valid_vendor_items else log.debug
                logger("product %s has %s vendorItems but each is missing (valid) vendor: %s",
                       product.upc, len(product.vendor_items), product)

        if vendor_items:
            if len(vendor_items) > 1:

                # try to narrow down a bit further, based on valid 'units' amount
                valid_items = [item for item in vendor_items
                               if item.units]
                if valid_items:
                    vendor_items = valid_items

                # warn if we still have more than one "obvious" vendor item
                if len(vendor_items) > 1:
                    logger = log.warning if self.warn_multiple_vendor_items else log.debug
                    logger("product %s has %s vendorItems to pick from: %s",
                           product.upc, len(vendor_items), product)

                    # sort the list so most recently modified is first
                    vendor_items.sort(key=lambda item: item.modified or self.old_datetime,
                                      reverse=True)

            # use the "first" vendor item available
            item = vendor_items[0]
            supplier_unit_id = item.sku
            supplier_id = item.vendor.name
            supplier_num_pkgs = item.units or 1
            if supplier_num_pkgs == 1:
                supplier_unit = 'LB' if product.scale else 'EA'
            else:
                supplier_unit = 'CS'

        pf1 = None
        if product.subdepartment:
            if not product.subdepartment.number:
                logger = log.warning if self.warn_empty_subdepartment else log.debug
                logger("product %s has 'empty' subdepartment number: %s",
                       product.upc, product)
            else:
                pf1 = "{} {}".format(product.subdepartment.number,
                                     product.subdepartment.name)

        memo = None
        if product.user_info and product.user_info.long_text is not None:
            memo = str(product.user_info.long_text)
        if memo and len(memo) > 254:
            logger = log.warning if self.warn_truncated_memo else log.debug
            logger("product %s has memo of length %s; will truncate: %s",
                   product.upc, len(memo), memo)
            memo = memo[:254]

        scale_ingredient_text = None
        if scale_item:
            scale_ingredient_text = scale_item.text
            if "\n" in scale_ingredient_text:
                logger = log.warning if self.warn_scale_ingredients_newline else log.debug
                logger("must remove carriage returns for scale ingredients: %s",
                       scale_ingredient_text)
                scale_ingredient_text = scale_ingredient_text.replace("\n", " ")

        return {
            '__product__': product,
            '__original_item_id__': product.upc,
            'uuid': get_uuid(),
            'item_id': item_id,
            'dept_id': department.number if department else None,
            'dept_name': department.name if department else None,
            'receipt_alias': product.description,
            'brand': product.brand,
            'item_name': product.description,
            'size': size,

            # TODO: does CORE have this?
            # 'sugg_retail': None,

            'last_cost': product.cost,
            'price_divider': price_divider,
            'base_price': product.normal_price,
            'ideal_margin': department.margin * 100 if department and department.margin else None,

            # TODO: does CORE have these?
            # 'disc_mult': None,

            'bottle_deposit': bottle_deposit,

            # TODO: does CORE have this?
            # 'pos_menu_group': None,

            'scale_label': scale_label,
            'sold_by_ea_or_lb': sold_by_ea_or_lb,
            'quantity_required': 'Y' if product.quantity_enforced else None,
            'weight_profile': weight_profile,
            'tax_1': tax_1 or None, # TODO: logic above is unfinished
            'tax_2': tax_2 or None, # TODO: logic above is unfinished
            'spec_tend_1': 'EBT' if product.foodstamp else None,
            'spec_tend_2': 'WIC' if product.wicable else None,
            'age_required': product.id_enforced or None,
            'location': location,

            # TODO: does CORE have these?
            # 'family_line': None,

            'alt_id': alt_id,
            'alt_receipt_alias': alt_receipt_alias,
            'alt_pkg_qty': alt_pkg_qty,
            'alt_pkg_price': alt_pkg_price,
            'auto_discount': auto_discount,
            'supplier_unit_id': supplier_unit_id,
            'supplier_id': supplier_id,
            'unit': supplier_unit,
            'num_pkgs': supplier_num_pkgs,

            # TODO: does CORE have these?
            # 'cs_pk_multiplier': None,
            # 'dsd': None,

            'pf1': pf1,

            # TODO: are these needed?
            # 'pf2',
            # 'pf3',
            # 'pf4',
            # 'pf5',
            # 'pf6',
            # 'pf7',
            # 'pf8',

            'memo': memo,
            'scale_shelf_life': scale_item.shelf_life if scale_item else None,
            'scale_shelf_life_type': 0 if scale_item else None,
            'scale_ingredient_text': scale_ingredient_text,
        }



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE Office -> CORE Lane import
"""

from corepos.db.office_op import Session as CoreOfficeSession, model as coreoffice
from corepos.db.lane_op import Session as CoreLaneSession

from rattail import importing
from rattail.importing.handlers import FromSQLAlchemyHandler, ToSQLAlchemyHandler
from rattail.util import OrderedDict
from rattail_corepos.corepos.lane.importing import op as corepos_importing


# TODO: this surely belongs in some other/common place? (is not lane-specific)
class FromCoreOfficeHandler(FromSQLAlchemyHandler):
    """
    Base class for import handlers which use CORE Office as the host.
    """
    host_title = "CORE Office"

    def make_host_session(self):
        return CoreOfficeSession()


# TODO: this surely belongs in some other/common place? (is not office-specific)
class ToCoreLaneHandler(ToSQLAlchemyHandler):
    """
    Base class for import handlers which target CORE Lane on the local side.
    """
    local_title = "CORE Lane"

    def make_session(self):
        return CoreLaneSession()


class FromCoreOfficeToCoreLane(FromCoreOfficeHandler, ToCoreLaneHandler):
    """
    Handler for CORE Office -> CORE Lane data export.
    """
    direction = 'export'
    dbkey = 'default'

    @property
    def local_title(self):
        return "CORE Lane ({})".format(self.dbkey)

    def make_session(self):
        return CoreLaneSession(bind=self.config.core_lane_op_engines[self.dbkey])

    def get_importers(self):
        importers = OrderedDict()
        importers['Department'] = DepartmentImporter
        importers['Product'] = ProductImporter
        return importers


class FromCore(importing.FromSQLAlchemy):
    """
    Base class for CORE Office -> CORE Lane data importers.
    """


class DepartmentImporter(FromCore, corepos_importing.model.DepartmentImporter):
    host_model_class = coreoffice.Department

    # these fields are held in common, between Office and Lane tables
    common_fields = [
        'number',
        'name',
        'tax',
        'food_stampable',
        'limit',
        'minimum',
        'discount',
        'see_id',
        'modified',
        'modified_by_id',
        'margin',
        'sales_code',
        'member_only',
        # TODO: these seem to be newer additions?
        # 'line_item_discount',
        # 'wicable',
    ]

    @property
    def supported_fields(self):
        return self.common_fields

    def normalize_host_object(self, department):
        data = dict([(field, getattr(department, field))
                     for field in self.common_fields])
        return data


class ProductImporter(FromCore, corepos_importing.model.ProductImporter):
    host_model_class = coreoffice.Product

    # these fields are held in common, between Office and Lane tables
    common_fields = [
        'id',
        'upc',
        'description',
        'brand',
        'formatted_name',
        'normal_price',
        'price_method',
        'group_price',
        'quantity',
        'special_price',
        'special_price_method',
        'special_group_price',
        'special_quantity',
        # 'special_limit',
        'start_date',
        'end_date',
        'department_number',
        'size',
        'tax_rate_id',
        'foodstamp',
        'scale',
        'scale_price',
        'mix_match_code',
        # 'created',
        # 'modified',

        # TODO: what to do about this 'replaces' thing?
        # 'batchID'=>array('type'=>'TINYINT', 'replaces'=>'advertised'),
        # batch_id = sa.Column('batchID', sa.SmallInteger(), nullable=True)
        # advertised = sa.Column(sa.Boolean(), nullable=True)

        'tare_weight',
        'discount',
        'discount_type',
        'line_item_discountable',
        'unit_of_measure',
        'wicable',
        'quantity_enforced',
        'id_enforced',
        'cost',
        # 'special_cost',
        # 'received_cost',
        'in_use',
        'flags',
        'subdepartment_number',
        'deposit',
        'local',
        'store_id',
        'default_vendor_id',
        'current_origin_id',
        # 'auto_par',
        # 'price_rule_id',
        'last_sold',
    ]

    @property
    def supported_fields(self):
        return self.common_fields

    def normalize_host_object(self, product):
        data = dict([(field, getattr(product, field))
                     for field in self.common_fields])
        return data



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE-POS commands
"""

from __future__ import unicode_literals, absolute_import

import sys

from rattail import commands
from rattail_corepos import __version__
from rattail.util import load_object


def main(*args):
    """
    Primary entry point for Crepes command system
    """
    if args:
        args = list(args)
    else:
        args = sys.argv[1:]

    cmd = Command()
    cmd.run(*args)


class Command(commands.Command):
    """
    Primary command for Crepes (CORE-POS)
    """
    name = 'crepes'
    version = __version__
    description = "Crepes -- command line interface for CORE-POS"
    long_description = ""


class ImportToCore(commands.ImportSubcommand):
    """
    Generic base class for commands which import *to* a CORE DB.
    """
    # subclass must set these!
    handler_key = None
    default_handler_spec = None

    def get_handler_factory(self, **kwargs):
        if self.config:
            spec = self.config.get('rattail.corepos.importing', '{}.handler'.format(self.handler_key),
                                   default=self.default_handler_spec)
        else:
            # just use default, for sake of cmd line help
            spec = self.default_handler_spec
        return load_object(spec)


class ExportCore(commands.ImportSubcommand):
    """
    Export data to another CORE database
    """
    name = 'export-core'
    description = __doc__.strip()
    default_handler_spec = 'rattail_corepos.corepos.importing.db.corepos:FromCoreToCoreExport'
    default_dbkey = 'host'

    def get_handler_factory(self, **kwargs):
        if self.config:
            spec = self.config.get('rattail_corepos.exporting', 'corepos.handler',
                                   default=self.default_handler_spec)
        else:
            # just use default, for sake of cmd line help
            spec = self.default_handler_spec
        return load_object(spec)

    def add_parser_args(self, parser):
        super(ExportCore, self).add_parser_args(parser)
        parser.add_argument('--dbkey', metavar='KEY', default=self.default_dbkey,
                            help="Config key for database engine to be used as the \"target\" "
                            "CORE DB, i.e. where data will be exported.  This key must be "
                            "defined in the [rattail_corepos.db] section of your config file.")

    def get_handler_kwargs(self, **kwargs):
        if 'args' in kwargs:
            kwargs['dbkey'] = kwargs['args'].dbkey
        return kwargs


class ExportCSV(commands.ExportFileSubcommand):
    """
    Export data from CORE to CSV file(s)
    """
    name = 'export-csv'
    description = __doc__.strip()
    default_handler_spec = 'rattail_corepos.corepos.importing.db.exporters.csv:FromCoreToCSV'

    def get_handler_factory(self, **kwargs):
        if self.config:
            spec = self.config.get('rattail_corepos.exporting', 'csv.handler',
                                   default=self.default_handler_spec)
        else:
            # just use default, for sake of cmd line help
            spec = self.default_handler_spec
        return load_object(spec)


class ImportCore(ImportToCore):
    """
    Import data from another CORE database
    """
    name = 'import-core'
    description = __doc__.strip()
    handler_key = 'corepos'
    default_handler_spec = 'rattail_corepos.corepos.importing.db.corepos:FromCoreToCoreImport'
    accepts_dbkey_param = True

    def add_parser_args(self, parser):
        super(ImportCore, self).add_parser_args(parser)
        if self.accepts_dbkey_param:
            parser.add_argument('--dbkey', metavar='KEY', default='host',
                                help="Config key for database engine to be used as the CORE "
                                "\"host\", i.e. the source of the data to be imported.  This key "
                                "must be defined in the [rattail_corepos.db] section of your config file.  "
                                "Defaults to 'host'.")

    def get_handler_kwargs(self, **kwargs):
        if self.accepts_dbkey_param:
            if 'args' in kwargs:
                kwargs['dbkey'] = kwargs['args'].dbkey
        return kwargs


class ImportCSV(commands.ImportFileSubcommand):
    """
    Import data from CSV file(s) to CORE database
    """
    name = 'import-csv'
    description = __doc__.strip()
    default_handler_spec = 'rattail_corepos.corepos.importing.db.csv:FromCSVToCore'

    def add_parser_args(self, parser):
        super(ImportCSV, self).add_parser_args(parser)
        parser.add_argument('--dbkey', metavar='KEY', default='default',
                            help="Config key for database engine to be used as the \"target\" "
                            "CORE DB, i.e. where data will be imported *to*.  This key must be "
                            "defined in the [rattail_corepos.db] section of your config file.")

    def get_handler_kwargs(self, **kwargs):
        kwargs = super(ImportCSV, self).get_handler_kwargs(**kwargs)
        if 'args' in kwargs:
            args = kwargs['args']
            kwargs['dbkey'] = args.dbkey
        return kwargs

    def get_handler_factory(self, **kwargs):
        if self.config:
            spec = self.config.get('rattail_corepos.importing', 'csv.handler',
                                   default=self.default_handler_spec)
        else:
            # just use default, for sake of cmd line help
            spec = self.default_handler_spec
        return load_object(spec)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE POS (DB) -> Rattail data importing
"""

import decimal

from corepos.db.office_op import model as corepos, Session as CoreSession

from rattail import importing
from rattail.gpc import GPC
from rattail.util import OrderedDict
from rattail_corepos import importing as corepos_importing


class FromCOREPOSToRattail(importing.FromSQLAlchemyHandler, importing.ToRattailHandler):
    """
    Import handler for data coming from a CORE POS database.
    """
    corepos_dbkey = 'default'

    @property
    def host_title(self):
        return "CORE-POS (DB/{})".format(self.corepos_dbkey)

    def make_host_session(self):
        return CoreSession(bind=self.config.corepos_engines[self.corepos_dbkey])

    def get_importers(self):
        importers = OrderedDict()
        importers['Vendor'] = VendorImporter
        importers['Department'] = DepartmentImporter
        importers['Subdepartment'] = SubdepartmentImporter
        importers['Product'] = ProductImporter
        return importers


class FromCOREPOS(importing.FromSQLAlchemy):
    """
    Base class for all CORE POS data importers.
    """


class VendorImporter(FromCOREPOS, corepos_importing.model.VendorImporter):
    """
    Importer for vendor data from CORE POS.
    """
    host_model_class = corepos.Vendor
    key = 'corepos_id'
    supported_fields = [
        'corepos_id',
        'name',
        'abbreviation',
        'special_discount',
        'phone_number',
        'fax_number',
        'email_address',
    ]

    def cache_query(self):
        """
        Return the query to be used when caching "local" data.
        """
        # can't just use rattail.db.model b/c the CoreVendor would normally not
        # be in there!  this still requires custom model to be configured though.
        model = self.config.get_model()

        # first get default query
        query = super(VendorImporter, self).cache_query()

        # maybe filter a bit, to ensure only "relevant" records are involved
        if 'corepos_id' in self.key:
            # note, the filter is probably redundant since we INNER JOIN on the
            # extension table, and it doesn't allow null ID values.  but clarity.
            query = query.join(model.CoreVendor)\
                         .filter(model.CoreVendor.corepos_id != None)

        return query

    def normalize_host_object(self, vendor):

        special_discount = None
        if vendor.discount_rate is not None:
            special_discount = decimal.Decimal('{:0.3f}'.format(vendor.discount_rate))

        return {
            'corepos_id': vendor.id,
            'name': vendor.name,
            'abbreviation': vendor.abbreviation or None,
            'special_discount': special_discount,
            'phone_number': vendor.phone or None,
            'fax_number': vendor.fax or None,
            'email_address': vendor.email or None,
        }


class DepartmentImporter(FromCOREPOS, corepos_importing.model.DepartmentImporter):
    """
    Importer for department data from CORE POS.
    """
    host_model_class = corepos.Department
    key = 'corepos_number'
    supported_fields = [
        'corepos_number',
        'number',
        'name',
    ]

    def normalize_host_object(self, department):
        return {
            'corepos_number': department.number,
            'number': department.number,
            'name': department.name,
        }


class SubdepartmentImporter(FromCOREPOS, corepos_importing.model.SubdepartmentImporter):
    """
    Importer for subdepartment data from CORE POS.
    """
    host_model_class = corepos.Subdepartment
    key = 'corepos_number'
    supported_fields = [
        'corepos_number',
        'number',
        'name',
        'department_number',
    ]

    def normalize_host_object(self, subdepartment):
        return {
            'corepos_number': subdepartment.number,
            'number': subdepartment.number,
            'name': subdepartment.name,
            'department_number': subdepartment.department_number,
        }


class ProductImporter(FromCOREPOS, corepos_importing.model.ProductImporter):
    """
    Importer for product data from CORE POS.
    """
    host_model_class = corepos.Product
    key = 'corepos_id'
    supported_fields = [
        'corepos_id',
        'item_id',
        'upc',
        'brand_name',
        'description',
        'size',
        'unit_size',
        'unit_of_measure',
        'uom_abbreviation',
        'weighed',
        'department_number',
        'subdepartment_number',
        'regular_price_price',
        'regular_price_multiple',
        'regular_price_type',
        'food_stampable',
        # 'tax1',
    ]

    def normalize_host_object(self, product):

        try:
            upc = GPC(product.upc, calc_check_digit='upc')
        except (TypeError, ValueError):
            log.debug("CORE POS product has invalid UPC: %s", product.upc)
            if len(self.key) == 1 and self.key[0] == 'upc':
                return
            upc = None

        price = None
        if product.normal_price is not None:
            price = decimal.Decimal('{:03f}'.format(product.normal_price))

        data = {
            'corepos_id': product.id,
            'item_id': product.upc,
            'upc': upc,
            'brand_name': (product.brand or '').strip() or None,
            'description': (product.description or '').strip(),

            'department_number': product.department_number or None,
            'subdepartment_number': product.subdepartment_number or None,

            'weighed': bool(product.scale),
            'food_stampable': product.foodstamp,
            # 'tax1': bool(product.tax), # TODO: is this right?

            'regular_price_price': price,
            'regular_price_multiple': 1 if price is not None else None,
            'regular_price_type': self.enum.PRICE_TYPE_REGULAR if price is not None else None,
        }

        if self.fields_active(self.size_fields):
            size_info = self.normalize_size_info(product)
            data.update({
                'size': size_info['size'],
                'unit_size': size_info['unit_size'],
                'unit_of_measure': size_info['uom_code'],
                'uom_abbreviation': size_info['uom_abbrev'],
            })

        return data



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE-POS misc. utilities
"""

import logging

import sqlalchemy as sa

from corepos.db.office_op import Session as CoreSession, model as corepos

from rattail.db.util import short_session
from rattail.util import OrderedDict, progress_loop


log = logging.getLogger(__name__)


def get_core_members(api, progress=None):
    """
    Shared logic for fetching *all* customer accounts from CORE-POS API.
    """
    # TODO: ideally could do this, but API doesn't let us fetch "all"
    # return api.get_members()

    # first we fetch all customer records from CORE DB
    with short_session(Session=CoreSession) as s:
        db_customers = s.query(corepos.Customer).all()
        s.expunge_all()

    # now we must fetch each customer account individually from API
    members = OrderedDict()

    def fetch(dbcust, i):
        if dbcust.card_number in members:
            return          # already fetched this one
        member = api.get_member(dbcust.card_number)
        if member:
            members[dbcust.card_number] = member
        else:
            logger = log.warning if dbcust.account_holder else log.debug
            logger("could not fetch member from CORE API: %s",
                   dbcust.card_number)

    progress_loop(fetch, db_customers, progress,
                  message="Fetching Member data from CORE API")
    return list(members.values())


def get_max_existing_vendor_id(session=None):
    """
    Returns the "last" (max) existing value for the ``vendors.vendorID``
    column, for use when creating new records, since it is not auto-increment.
    """
    with short_session(Session=CoreSession, session=session) as s:
        return s.query(sa.func.max(corepos.Vendor.id))\
                .scalar() or 0



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE-POS model importers (direct DB)

.. warning::
   All classes in this module are "direct DB" importers, which will write
   directly to MySQL.  They are meant to be used in dry-run mode only, and/or
   for sample data import to a dev system etc.  They are *NOT* meant for
   production use, as they will completely bypass any CORE business rules logic
   which may exist.
"""

from sqlalchemy.orm.exc import NoResultFound

from rattail import importing

from corepos.db.office_op import model as corepos
from corepos.db.office_trans import model as coretrans


class ToCore(importing.ToSQLAlchemy):
    """
    Base class for all CORE "operational" model importers.
    """

    def create_object(self, key, host_data):

        # NOTE! some tables in CORE DB may be using the MyISAM storage engine,
        # which means it is *not* transaction-safe and therefore we cannot rely
        # on "rollback" if in dry-run mode!  in other words we better not touch
        # the record at all, for dry run
        if self.dry_run:
            return host_data

        return super(ToCore, self).create_object(key, host_data)

    def update_object(self, obj, host_data, **kwargs):

        # NOTE! some tables in CORE DB may be using the MyISAM storage engine,
        # which means it is *not* transaction-safe and therefore we cannot rely
        # on "rollback" if in dry-run mode!  in other words we better not touch
        # the record at all, for dry run
        if self.dry_run:
            return obj

        return super(ToCore, self).update_object(obj, host_data, **kwargs)

    def delete_object(self, obj):

        # NOTE! some tables in CORE DB may be using the MyISAM storage engine,
        # which means it is *not* transaction-safe and therefore we cannot rely
        # on "rollback" if in dry-run mode!  in other words we better not touch
        # the record at all, for dry run
        if self.dry_run:
            return True

        return super(ToCore, self).delete_object(obj)


class ToCoreTrans(importing.ToSQLAlchemy):
    """
    Base class for all CORE "transaction" model importers
    """


########################################
# CORE Operational
########################################

class DepartmentImporter(ToCore):
    model_class = corepos.Department
    key = 'number'


class SubdepartmentImporter(ToCore):
    model_class = corepos.Subdepartment
    key = 'number'


class VendorImporter(ToCore):
    model_class = corepos.Vendor
    key = 'id'


class VendorContactImporter(ToCore):
    model_class = corepos.VendorContact
    key = 'vendor_id'


class ProductImporter(ToCore):
    model_class = corepos.Product
    key = 'id'


class ProductFlagImporter(ToCore):
    model_class = corepos.ProductFlag
    key = 'bit_number'


class VendorItemImporter(ToCore):
    model_class = corepos.VendorItem
    key = ('sku', 'vendor_id')


class EmployeeImporter(ToCore):
    model_class = corepos.Employee
    key = 'number'


class CustDataImporter(ToCore):
    model_class = corepos.CustData
    key = 'id'


class MemberTypeImporter(ToCore):
    model_class = corepos.MemberType
    key = 'id'


class MemberInfoImporter(ToCore):
    model_class = corepos.MemberInfo
    key = 'card_number'


class MemberDateImporter(ToCore):
    model_class = corepos.MemberDate
    key = 'card_number'


class MemberContactImporter(ToCore):
    model_class = corepos.MemberContact
    key = 'card_number'


class HouseCouponImporter(ToCore):
    model_class = corepos.HouseCoupon
    key = 'coupon_id'


########################################
# CORE Transactions
########################################

class TransactionDetailImporter(ToCoreTrans):
    """
    CORE-POS transaction data importer.
    """
    model_class = coretrans.TransactionDetail



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE-POS -> Catapult Membership Workbook
"""

import decimal
import logging

from sqlalchemy import orm

from corepos.db.office_op import model as corepos

from rattail.util import OrderedDict
from rattail.importing.handlers import ToFileHandler
from rattail.time import localtime
from rattail.excel import ExcelReader
from rattail_corepos.corepos.importing.db.corepos import FromCoreHandler, FromCore
from rattail_onager.catapult.importing import membership as catapult_importing


log = logging.getLogger(__name__)


class FromCoreToCatapult(FromCoreHandler, ToFileHandler):
    """
    Handler for CORE -> Catapult (Membership Workbook)
    """
    host_title = "CORE-POS"
    local_title = "Catapult (Membership Workbook)"
    direction = 'export'

    def get_importers(self):
        importers = OrderedDict()
        importers['Member'] = MemberImporter
        return importers


class MemberImporter(FromCore, catapult_importing.model.MemberImporter):
    """
    Member data importer.
    """
    host_model_class = corepos.CustData
    supported_fields = [
        'member_id',
        'first_name',
        'last_name',
        'address_1',
        'address_2',
        'city',
        'state',
        'zip_code',
        'phone_number',
        'email',
        'member_notes',
        'member_join_date',
        'family_affiliation',
        'account_number',
        'membership_profile_name',
    ]

    def setup(self):
        super(MemberImporter, self).setup()

        self.warn_truncated_field = self.config.getbool(
            'corepos', 'exporting.catapult_membership.warn_truncated_field',
            default=True)

        self.cache_membership_profiles()

    def cache_membership_profiles(self):
        self.membership_profiles = {}
        sheet_name = self.config.get('catapult', 'membership_profiles_worksheet_name',
                                     default="Membership Profile details")
        reader = ExcelReader(self.workbook_template_path, sheet_name=sheet_name)
        for profile in reader.read_rows(progress=self.progress):
            if profile['Equity Paid in Full Amount']:
                profile['Equity Paid in Full Amount'] = decimal.Decimal(profile['Equity Paid in Full Amount'])
            self.membership_profiles[profile['Membership Profile Name']] = profile

        # also figure out which profile is default
        self.default_membership_profile = None
        for profile in self.membership_profiles.values():
            if profile['Please indicate Default Profile'] == 'X':
                self.default_membership_profile = profile
                break
        if not self.default_membership_profile:
            raise RuntimeError("cannot determine default membership profile")

    def query(self):
        return self.host_session.query(corepos.CustData)\
                                .order_by(corepos.CustData.card_number,
                                          corepos.CustData.person_number,
                                          corepos.CustData.id)\
                                .options(orm.joinedload(corepos.CustData.member_type))\
                                .options(orm.joinedload(corepos.CustData.member_info)\
                                         .joinedload(corepos.MemberInfo.dates))\
                                .options(orm.joinedload(corepos.CustData.member_info)\
                                         .joinedload(corepos.MemberInfo.notes))

    def normalize_host_object(self, custdata):

        if custdata.person_number == 1:
            family_affiliation = False
        elif custdata.person_number > 1:
            family_affiliation = True
        else:
            log.warning("member #%s has unexpected person_number (%s): %s",
                        custdata.card_number, custdata.person_number, custdata)
            family_affiliation = False

        if custdata.member_type:
            membership_profile_name = custdata.member_type.description
        else:
            log.warning("member #%s has no member type: %s",
                        custdata.card_number, custdata)
            membership_profile_name = self.default_membership_profile['Membership Profile Name']

        data = {
            'member_id': str(custdata.id),
            'first_name': custdata.first_name,
            'last_name': custdata.last_name,

            # these will be blank unless we have an associated `meminfo` record
            'phone_number': None,
            'email': None,
            'address_1': None,
            'address_2': None,
            'city': None,
            'state': None,
            'zip_code': None,
            'member_join_date': None,
            'member_notes': None,

            'family_affiliation': family_affiliation,
            'account_number': str(custdata.card_number),
            'membership_profile_name': membership_profile_name,
        }

        info = custdata.member_info
        if info:
            data['phone_number'] = info.phone
            data['email'] = info.email
            data['address_1'], data['address_2'] = info.split_street()
            data['city'] = info.city
            data['state'] = info.state
            data['zip_code'] = info.zip

            if info.dates:
                if len(info.dates) > 1:
                    log.warning("member #%s has multiple (%s) `memDates` records: %s",
                                custdata.card_number, len(info.dates), custdata)
                dates = info.dates[0]
                if dates.start_date:
                    start_date = localtime(self.config, dates.start_date).date()
                    data['member_join_date'] = start_date.strftime('%Y-%m-%d')

            if info.notes:
                notes = []
                for note in reversed(info.notes): # show most recent first
                    text = str(note.note or '').strip() or None
                    if text:
                        notes.append(text)
                if notes:
                    data['member_notes'] = '\n'.join(notes)

        return data



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE-POS -> Trainwreck data importing
"""

from corepos.db.office_trans import Session as CoreTransSession, model as coretrans
from corepos.db.office_op import Session as CoreSession, model as corepos

from rattail import importing
from rattail.util import OrderedDict
from rattail.time import localtime, make_utc
from rattail.trainwreck import importing as trainwreck_importing


class FromCoreToTrainwreck(importing.FromSQLAlchemyHandler, trainwreck_importing.ToTrainwreckHandler):
    """
    Import data from CORE-POS into Trainwreck
    """
    host_title = "CORE-POS"
    corepos_dbkey = 'default'

    def make_host_session(self):
        return CoreTransSession(bind=self.config.coretrans_engines[self.corepos_dbkey])

    def get_importer_kwargs(self, key, **kwargs):
        kwargs = super(FromCoreToTrainwreck, self).get_importer_kwargs(key, **kwargs)
        kwargs.setdefault('core_op_session', self.core_op_session)
        return kwargs

    def begin_host_transaction(self):
        super(FromCoreToTrainwreck, self).begin_host_transaction()

        self.core_op_session = CoreSession(bind=self.config.corepos_engines[self.corepos_dbkey])

    def rollback_host_transaction(self):
        super(FromCoreToTrainwreck, self).rollback_host_transaction()

        self.core_op_session.rollback()
        self.core_op_session.close()
        self.core_op_session = None

    def commit_host_transaction(self):
        super(FromCoreToTrainwreck, self).commit_host_transaction()

        self.core_op_session.commit()
        self.core_op_session.close()
        self.core_op_session = None

    def get_importers(self):
        importers = OrderedDict()
        importers['Transaction'] = TransactionImporter
        importers['TransactionItem'] = TransactionItemImporter
        return importers


class FromCore(importing.Importer):
    """
    Base class for CORE importers.
    """

    def fetch_details(self):
        # TODO: should figure out "which" txns apply to our date range first,
        # w/ a rather "loose" query, to avoid issues when a txn spans multiple
        # dates..?
        return self.host_session.query(coretrans.TransactionDetail)\
                                .filter(coretrans.TransactionDetail.date_time >= self.start_time)\
                                .filter(coretrans.TransactionDetail.date_time < self.end_time)\
                                .order_by(coretrans.TransactionDetail.register_number,
                                          coretrans.TransactionDetail.transaction_number,
                                          coretrans.TransactionDetail.store_row_id)\
                                .all()

    def make_system_id(self, detail):
        assert detail.employee_number
        core_id = '-'.join([
            str(detail.employee_number),
            str(detail.register_number),
            str(detail.transaction_number),
        ])
        return '|'.join([
            str(detail.store_id),
            core_id,
        ])


class TransactionImporter(FromCore, trainwreck_importing.model.TransactionImporter):
    """
    Import transaction data from CORE-POS
    """
    key = ('system', 'system_id')
    importing_from_system = 'corepos'
    supported_fields = [
        'system',
        'system_id',
        'terminal_id',
        'receipt_number',
        'start_time',
        'end_time',
        'cashier_id',
        'cashier_name',
        'customer_id',
        'customer_name',
        'subtotal',
        'total',
    ]

    def setup(self):
        super(TransactionImporter, self).setup()

        app = self.config.get_app()
        self.people_handler = app.get_people_handler()

        if 'cashier_name' in self.fields:
            self.corepos_employees = self.cache_model(corepos.Employee,
                                                      session=self.core_op_session,
                                                      key='number')

        if 'customer_name' in self.fields:
            query = self.core_op_session.query(corepos.CustData)\
                                        .filter(corepos.CustData.person_number == 1)
            self.corepos_customers = self.cache_model(corepos.CustData,
                                                      session=self.core_op_session,
                                                      query=query,
                                                      key='card_number')

    def get_host_objects(self):
        details = self.fetch_details()
        transactions = []
        current = {}

        def collect(detail, i):
            receipt_number = str(detail.transaction_number)

            date_time = detail.date_time
            if date_time:
                date_time = localtime(self.config, date_time)
                date_time = make_utc(date_time)

            if current and current['receipt_number'] != receipt_number:
                transactions.append(dict(current))
                current.clear()

            if not current:
                current.update({
                    'system': self.enum.TRAINWRECK_SYSTEM_COREPOS,
                    'system_id': self.make_system_id(detail),
                    'terminal_id': str(detail.register_number),
                    'receipt_number': receipt_number,
                    'cashier_id': str(detail.employee_number) if detail.employee_number else None,
                    'customer_id': str(detail.card_number) if detail.card_number else None,
                    'start_time': date_time,
                    'end_time': date_time,
                })

            if detail.transaction_type == 'C':
                if 'Subtotal' in detail.description:
                    current['subtotal'] = detail.unit_price
                    current['total'] = detail.unit_price

            if date_time:
                current['end_time'] = date_time

        self.progress_loop(collect, details,
                           message="Collecting transaction data")

        # don't forget to add the last one!
        if current:
            transactions.append(current)
        return transactions

    def normalize_host_object(self, txn):

        if 'cashier_name' in self.fields:
            txn['cashier_name'] = None
            if txn['cashier_id']:
                employee = self.corepos_employees.get(int(txn['cashier_id']))
                if employee:
                    txn['cashier_name'] = self.people_handler.normalize_full_name(
                        employee.first_name, employee.last_name)

        if 'customer_name' in self.fields:
            txn['customer_name'] = None
            if txn['customer_id']:
                custdata = self.corepos_customers.get(int(txn['customer_id']))
                if custdata:
                    txn['customer_name'] = self.people_handler.normalize_full_name(
                        custdata.first_name, custdata.last_name)

        return txn


class TransactionItemImporter(FromCore, trainwreck_importing.model.TransactionItemImporter):
    """
    Import transaction item data from CORE-POS
    """
    key = ('transaction_system_id', 'sequence')
    importing_from_system = 'corepos'
    supported_fields = [
        'transaction_system_id',
        'sequence',
        'item_scancode',
        'department_number',
        'description',
        'unit_price',
        'unit_quantity',
        'total',
        'void',
    ]

    def get_host_objects(self):
        return self.fetch_details()

    def normalize_host_object(self, detail):

        # TODO: this needs to be a lot smarter.  for now this "works" i guess
        if detail.transaction_type != 'I':
            return

        return {
            'transaction_system_id': self.make_system_id(detail),
            'sequence': detail.transaction_id,
            'item_scancode': detail.upc,
            'department_number': detail.department_number,
            'description': detail.description,
            'unit_price': detail.unit_price,
            'unit_quantity': detail.quantity,
            'total': detail.total,
            'void': bool(detail.voided),
        }



# -*- coding: utf-8 -*-
"""initial CORE-POS integration tables

Revision ID: b43e93d32275
Revises: dfc1ed686f3f
Create Date: 2020-03-04 14:21:23.625568

"""

# revision identifiers, used by Alembic.
revision = 'b43e93d32275'
down_revision = None
branch_labels = ('rattail_corepos',)
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # corepos_vendor
    op.create_table('corepos_vendor',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('corepos_id', sa.Integer(), nullable=False),
                    sa.ForeignKeyConstraint(['uuid'], ['vendor.uuid'], name='corepos_vendor_fk_vendor'),
                    sa.PrimaryKeyConstraint('uuid')
    )
    op.create_table('corepos_vendor_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('corepos_id', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_corepos_vendor_version_end_transaction_id'), 'corepos_vendor_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_corepos_vendor_version_operation_type'), 'corepos_vendor_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_corepos_vendor_version_transaction_id'), 'corepos_vendor_version', ['transaction_id'], unique=False)


def downgrade():

    # corepos_vendor
    op.drop_index(op.f('ix_corepos_vendor_version_transaction_id'), table_name='corepos_vendor_version')
    op.drop_index(op.f('ix_corepos_vendor_version_operation_type'), table_name='corepos_vendor_version')
    op.drop_index(op.f('ix_corepos_vendor_version_end_transaction_id'), table_name='corepos_vendor_version')
    op.drop_table('corepos_vendor_version')
    op.drop_table('corepos_vendor')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
"operational" model importers for CORE Lane

.. warning::
   All classes in this module are "direct DB" importers, which will
   write directly to MySQL.  They are meant to be used in dry-run mode
   only, and/or for sample data import to a dev system etc.  They are
   *NOT* meant for production use, as they will completely bypass any
   CORE business rules logic which may exist.
"""

from rattail import importing
from rattail_corepos.corepos.common.importing import ToCore

from corepos.db.lane_op import model as corepos


class DepartmentImporter(ToCore):
    model_class = corepos.Department
    key = 'number'


class ProductImporter(ToCore):
    model_class = corepos.Product
    key = 'id'



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Rattail-COREPOS Config Extension
"""

from rattail.config import ConfigExtension
from rattail.db.config import get_engines


class RattailCOREPOSExtension(ConfigExtension):
    """
    Config extension for Rattail-COREPOS
    """
    key = 'rattail-corepos'

    def configure(self, config):

        # office_op
        from corepos.db.office_op import Session
        engines = get_engines(config, section='corepos.db.office_op')
        config.core_office_op_engines = engines
        config.core_office_op_engine = engines.get('default')
        Session.configure(bind=config.core_office_op_engine)
        # TODO: deprecate / remove these next 2 lines
        config.corepos_engines = engines
        config.corepos_engine = engines.get('default')

        # office_trans
        from corepos.db.office_trans import Session
        engines = get_engines(config, section='corepos.db.office_trans')
        config.core_office_trans_engines = engines
        config.core_office_trans_engine = engines.get('default')
        Session.configure(bind=config.core_office_trans_engine)
        # TODO: deprecate / remove these next 2 lines
        config.coretrans_engines = engines
        config.coretrans_engine = engines.get('default')

        # lane_op
        from corepos.db.lane_op import Session
        engines = get_engines(config, section='corepos.db.lane_op')
        config.core_lane_op_engines = engines
        config.core_lane_op_engine = engines.get('default')
        Session.configure(bind=config.core_lane_op_engine)


def core_office_url(config, require=False, **kwargs):
    """
    Returns the base URL for the CORE Office web app.  Note that this URL will
    *not* have a trailing slash.
    """
    args = ['corepos', 'office.url']
    if require:
        url = config.require(*args, **kwargs)
        return url.rstrip('/')
    else:
        url = config.get(*args, **kwargs)
        if url:
            return url.rstrip('/')


def core_office_customer_account_url(config, card_number, office_url=None):
    """
    Returns the CORE Office URL for the customer account with the given card
    number.
    """
    if not office_url:
        office_url = core_office_url(config, require=True)
    return '{}/mem/MemberEditor.php?memNum={}'.format(
        office_url, card_number)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Handler for Vendor Catalog batches
"""

import decimal

from rattail.batch import vendorcatalog as base
from rattail_corepos.corepos.api import make_corepos_api


class VendorCatalogHandler(base.VendorCatalogHandler):
    """
    Handler for vendor catalog batches.
    """
    # default logic tries to update versions after execution, but since we just
    # update CORE-POS via API there's no need
    version_catchup_execute = None

    def add_row(self, batch, row):

        # CORE unit costs seem to be rounded to 3 decimal places
        if row.unit_cost:
            row.unit_cost = row.unit_cost.quantize(decimal.Decimal('0.123'))

        # CORE bases its cost off the unit instead of case.  so here we make
        # sure the case cost is cleanly derived from unit cost.
        if row.case_cost and row.unit_cost and row.case_size:
            row.case_cost = row.unit_cost * row.case_size

        # okay now continue as normal
        super(VendorCatalogHandler, self).add_row(batch, row)

    def describe_execution(self, batch, **kwargs):
        return ("The `vendorItems` table in CORE-POS will be updated directly "
                "via API, for all rows indicating a change etc.  You may also "
                "update `products.cost` if desired.")

    def execute(self, batch, progress=None, **kwargs):
        """
        Update CORE-POS etc.
        """
        rows = [row for row in batch.active_rows()
                if row.status_code in (row.STATUS_NEW_COST,
                                       row.STATUS_UPDATE_COST,
                                       row.STATUS_CHANGE_VENDOR_ITEM_CODE,
                                       row.STATUS_CHANGE_CASE_SIZE,
                                       row.STATUS_CHANGE_COST)]

        if rows:
            vendor_id = batch.vendor.corepos_id
            if not vendor_id:
                raise ValueError("Batch vendor does not have valid CORE-POS ID")

            self.api = make_corepos_api(self.config)
            self.update_corepos(batch, rows, vendor_id, progress=progress,
                                # update_product_costs=kwargs.get('update_product_costs', False),
            )

        return True

    def update_corepos(self, batch, rows, vendor_id, progress=None,
                       # TODO: this kwarg seems perhaps useful, but for now we
                       # are auto-detecting when such an update is needed
                       #update_product_costs=False,
    ):
        """
        Update the `vendorItems` table in CORE-POS (and maybe `products` too).
        """
        def update(row, i):
            # we may need this value in a couple places
            unit_cost = float(row.unit_cost)

            # we will want to set default vendor for the product, if it does
            # not yet have one
            first_vendor = False
            if not row.product.costs:
                first_vendor = True
            # core_product = self.api.get_product(row.item_id)
            # if not core_product['default_vendor_id']:
            #     first_vendor = True

            # figure out if we are "updating the same, primary" cost record,
            # b/c if so we will want to update product accordingly also.  this
            # is always the case when this is the first vendor for product.
            updating_primary = first_vendor
            if not updating_primary:
                cost = row.product.cost
                if cost and cost is row.cost:
                    updating_primary = True
                # core_items = self.api.get_vendor_items(upc=row.item_id)
                # if core_items:
                #     core_item = core_items[0]
                #     if core_item['sku'] == row.vendor_code:
                #         updating_primary = True

            # create or update the `vendorItems` record in CORE
            self.api.set_vendor_item(sku=row.vendor_code,
                                     vendorID=vendor_id,
                                     upc=row.item_id,
                                     brand=row.brand_name,
                                     description=row.description,
                                     size=row.size,
                                     units=row.case_size,
                                     cost=unit_cost,
                                     # TODO: we (may) have vendor SRP, but pretty
                                     # sure CORE has different plans for its `srp`
                                     #srp=row.suggested_retail,
            )

            # TODO: CORE does not have the concept of a true "default"
            # `vendorItems` record, but rather uses the `modified` timestamp
            # for pseudo-default.  this means any given product may wind up
            # with a new/different pseudo-default when the above operation
            # completes.  in which case, perhaps we should *always* update
            # `products.cost` accordingly (below)..?  er, still only if the
            # product's `default_vendor_id` matches at least, i guess...  for
            # now we are only doing so if it "obviously" needs it.

            # maybe also update `products` record in CORE
            if updating_primary:
                kwargs = {'cost': unit_cost}
                if first_vendor:
                    kwargs['default_vendor_id'] = vendor_id
                self.api.set_product(upc=row.item_id, **kwargs)

        self.progress_loop(update, rows, progress,
                           message="Updating CORE-POS via API")



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Rattail -> Rattail "versions" data import
"""

from rattail.importing import versions as base


class CoreposVersionMixin(object):
    """
    Add default registration of custom importers
    """

    def add_corepos_importers(self, importers):
        importers['CorePerson'] = CorePersonImporter
        importers['CoreCustomer'] = CoreCustomerImporter
        importers['CoreMember'] = CoreMemberImporter
        importers['CoreStore'] = CoreStoreImporter
        importers['CoreDepartment'] = CoreDepartmentImporter
        importers['CoreSubdepartment'] = CoreSubdepartmentImporter
        importers['CoreVendor'] = CoreVendorImporter
        importers['CoreProduct'] = CoreProductImporter
        return importers


class CorePersonImporter(base.VersionImporter):

    @property
    def host_model_class(self):
        model = self.config.get_model()
        return model.CorePerson


class CoreCustomerImporter(base.VersionImporter):

    @property
    def host_model_class(self):
        model = self.config.get_model()
        return model.CoreCustomer


class CoreMemberImporter(base.VersionImporter):

    @property
    def host_model_class(self):
        model = self.config.get_model()
        return model.CoreMember


class CoreStoreImporter(base.VersionImporter):

    @property
    def host_model_class(self):
        model = self.config.get_model()
        return model.CoreStore


class CoreDepartmentImporter(base.VersionImporter):

    @property
    def host_model_class(self):
        model = self.config.get_model()
        return model.CoreDepartment


class CoreSubdepartmentImporter(base.VersionImporter):

    @property
    def host_model_class(self):
        model = self.config.get_model()
        return model.CoreSubdepartment


class CoreVendorImporter(base.VersionImporter):

    @property
    def host_model_class(self):
        model = self.config.get_model()
        return model.CoreVendor


class CoreProductImporter(base.VersionImporter):

    @property
    def host_model_class(self):
        model = self.config.get_model()
        return model.CoreProduct



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE POS (API) -> Rattail data importing
"""

import datetime
import decimal
import logging

from sqlalchemy import orm

from rattail import importing
from rattail.gpc import GPC
from rattail.util import OrderedDict
from rattail.time import localtime, make_utc
from rattail.core import get_uuid
from rattail.db.util import normalize_full_name
from rattail_corepos import importing as corepos_importing
from rattail_corepos.corepos.util import get_core_members
from rattail_corepos.corepos.api import make_corepos_api


log = logging.getLogger(__name__)


class FromCOREPOSToRattail(importing.ToRattailHandler):
    """
    Import handler for data coming from a CORE POS API.
    """
    host_title = "CORE-POS (API)"

    def get_importers(self):
        importers = OrderedDict()
        importers['Customer'] = CustomerImporter
        importers['Person'] = PersonImporter
        importers['CustomerPerson'] = CustomerPersonImporter
        importers['Member'] = MemberImporter
        importers['Store'] = StoreImporter
        importers['Department'] = DepartmentImporter
        importers['Subdepartment'] = SubdepartmentImporter
        importers['Vendor'] = VendorImporter
        importers['Product'] = ProductImporter
        importers['ProductCost'] = ProductCostImporter
        importers['ProductMovement'] = ProductMovementImporter
        return importers

    def get_default_keys(self):
        keys = super(FromCOREPOSToRattail, self).get_default_keys()
        if 'ProductMovement' in keys:
            keys.remove('ProductMovement')
        return keys


class FromCOREPOSAPI(importing.Importer):
    """
    Base class for all CORE POS API data importers.
    """

    def setup(self):
        super(FromCOREPOSAPI, self).setup()
        self.establish_api()

    def establish_api(self):
        self.api = make_corepos_api(self.config)

    def get_core_members(self):
        return get_core_members(self.api, progress=self.progress)


class CustomerImporter(FromCOREPOSAPI, corepos_importing.model.CustomerImporter):
    """
    Importer for customer data from CORE POS API.
    """
    key = 'corepos_account_id'
    supported_fields = [
        'corepos_account_id',
        'id',
        'number',
        'name',
        'address_street',
        'address_street2',
        'address_city',
        'address_state',
        'address_zipcode',
    ]

    def get_host_objects(self):
        return self.get_core_members()

    def normalize_host_object(self, member):

        if member['customerAccountID'] == 0:
            log.debug("member %s has customerAccountID of 0: %s",
                      member['cardNo'], member)
            return

        # figure out the "account holder" customer for the member.  note that
        # we only use this to determine the `Customer.name` in Rattail
        customers = member['customers']
        account_holders = [customer for customer in customers
                           if customer['accountHolder']]
        if account_holders:
            if len(account_holders) > 1:
                log.warning("member %s has %s account holders in CORE: %s",
                            member['cardNo'], len(account_holders), member)
            customer = account_holders[0]
        elif customers:
            if len(customers) > 1:
                log.warning("member %s has %s customers but no account holders: %s",
                            member['cardNo'], len(customers), member)
            customer = customers[0]
        else:
            raise NotImplementedError("TODO: how to handle member with no customers?")

        return {
            'corepos_account_id': int(member['customerAccountID']),
            'id': member['customerAccountID'],
            'number': int(member['cardNo']),
            'name': normalize_full_name(customer['firstName'],
                                        customer['lastName']),

            'address_street': member['addressFirstLine'] or None,
            'address_street2': member['addressSecondLine'] or None,
            'address_city': member['city'] or None,
            'address_state': member['state'] or None,
            'address_zipcode': member['zip'] or None,
        }


class PersonImporter(FromCOREPOSAPI, corepos_importing.model.PersonImporter):
    """
    Importer for person data from CORE POS API.
    """
    key = 'corepos_customer_id'
    supported_fields = [
        'corepos_customer_id',
        'first_name',
        'last_name',
        'display_name',
        'customer_uuid',
        'customer_person_ordinal',
        'phone_number',
        'phone_number_2',
        'email_address',
    ]

    def setup(self):
        super(PersonImporter, self).setup()
        model = self.config.get_model()

        self.customers = self.cache_model(model.Customer, key='id')

    def get_host_objects(self):

        # first get all member data from CORE API
        members = self.get_core_members()
        normalized = []

        # then collect all the "person" records
        def normalize(member, i):
            normalized.extend(self.get_person_objects_for_member(member))

        self.progress_loop(normalize, members,
                           message="Collecting Person data from CORE")
        return normalized

    def get_person_objects_for_member(self, member):
        """
        Return a list of Person data objects for the given Member.  This
        logic is split out separately so that datasync can leverage it too.
        """
        customers = member['customers']
        people = []

        # make sure account holder is listed first
        account_holder = None
        secondary = False
        mixedup = False
        for customer in customers:
            if customer['accountHolder'] and not secondary:
                account_holder = customer
            elif not customer['accountHolder']:
                secondary = True
            elif customer['accountHolder'] and secondary:
                mixedup = True
        if mixedup:
            raise NotImplementedError("TODO: should re-sort the customers list for member {}".format(member['cardNo']))

        for i, customer in enumerate(customers, 1):
            person = dict(customer)
            person['customer_person_ordinal'] = i
            people.append(person)

        return people

    def get_customer(self, id):
        if hasattr(self, 'customers'):
            return self.customers.get(id)

        model = self.config.get_model()
        try:
            return self.session.query(model.Customer)\
                               .filter(model.Customer.id == id)\
                               .one()
        except orm.exc.NoResultFound:
            pass

    def normalize_host_object(self, person):

        customer = self.get_customer(person['customerAccountID'])
        if not customer:
            log.warning("Rattail customer not found for customerAccountID: %s",
                        person['customerAccountID'])
            return

        return {
            'corepos_customer_id': int(person['customerID']),
            'first_name': person['firstName'],
            'last_name': person['lastName'],
            'display_name': normalize_full_name(person['firstName'],
                                                person['lastName']),
            'customer_uuid': customer.uuid,
            'customer_person_ordinal': person['customer_person_ordinal'],
            'phone_number': person['phone'] or None,
            'phone_number_2': person['altPhone'] or None,
            'email_address': person['email'] or None,
        }


class CustomerPersonImporter(FromCOREPOSAPI, importing.model.CustomerPersonImporter):
    """
    Importer for customer-person linkage data from CORE POS API.

    Note that we don't use this one in datasync, it's just for nightly
    double-check.
    """
    key = ('customer_uuid', 'person_uuid')
    supported_fields = [
        'customer_uuid',
        'person_uuid',
        'ordinal',
    ]

    def setup(self):
        super(CustomerPersonImporter, self).setup()
        model = self.config.get_model()

        query = self.session.query(model.Customer)\
                            .join(model.CoreCustomer)
        self.customers = self.cache_model(model.Customer, query=query,
                                          key='corepos_account_id')

        query = self.session.query(model.Person)\
                            .join(model.CorePerson)\
                            .filter(model.CorePerson.corepos_customer_id != None)
        self.people = self.cache_model(model.Person, query=query,
                                       key='corepos_customer_id',
                                       query_options=[orm.joinedload(model.Person._corepos)])

    def get_host_objects(self):

        # first get all member data from CORE API
        members = self.get_core_members()
        normalized = []

        # then collect all customer/person combination records
        def normalize(member, i):
            # make sure we put the account holder first in the list!
            customers = sorted(member['customers'],
                               key=lambda cust: 1 if cust['accountHolder'] else 0,
                               reverse=True)
            for i, customer in enumerate(customers, 1):
                normalized.append({
                    'customer_account_id': int(member['customerAccountID']),
                    'person_customer_id': customer['customerID'],
                    'ordinal': i,
                })

        self.progress_loop(normalize, members,
                           message="Collecting CustomerPerson data from CORE")
        return normalized

    def get_customer(self, account_id):
        if hasattr(self, 'customers'):
            return self.customers.get(account_id)

        model = self.config.get_model()
        try:
            return self.session.query(model.Customer)\
                               .join(model.CoreCustomer)\
                               .filter(model.CoreCustomer.corepos_account_id == account_id)\
                               .one()
        except orm.exc.NoResultFound:
            pass

    def get_person(self, corepos_customer_id):
        if hasattr(self, 'people'):
            return self.people.get(corepos_customer_id)

        model = self.config.get_model()
        try:
            return self.session.query(model.Person)\
                               .join(model.CorePerson)\
                               .filter(model.CorePerson.corepos_customer_id == corepos_customer_id)\
                               .one()
        except orm.exc.NoResultFound:
            pass

    def normalize_host_object(self, cp):

        customer = self.get_customer(cp['customer_account_id'])
        if not customer:
            log.warning("Rattail customer not found for customerAccountID: %s",
                        cp['customer_account_id'])
            return

        person = self.get_person(int(cp['person_customer_id']))
        if not person:
            log.warning("Rattail person not found for customerID: %s",
                        cp['person_customer_id'])
            return

        return {
            'customer_uuid': customer.uuid,
            'person_uuid': person.uuid,
            'ordinal': cp['ordinal'],
        }


class StoreImporter(FromCOREPOSAPI, corepos_importing.model.StoreImporter):
    """
    Importer for store data from CORE POS API.
    """
    key = 'corepos_id'
    supported_fields = [
        'corepos_id',
        'id',
        'name',
    ]

    def get_host_objects(self):
        return self.api.get_stores()

    def normalize_host_object(self, store):
        return {
            'corepos_id': int(store['storeID']),
            'id': str(store['storeID']),
            'name': store['description'],
        }


class DepartmentImporter(FromCOREPOSAPI, corepos_importing.model.DepartmentImporter):
    """
    Importer for department data from CORE POS API.
    """
    key = 'corepos_number'
    supported_fields = [
        'corepos_number',
        'number',
        'name',
    ]

    def get_host_objects(self):
        return self.api.get_departments()

    def normalize_host_object(self, department):
        return {
            'corepos_number': int(department['dept_no']),
            'number': int(department['dept_no']),
            'name': department['dept_name'],
        }


class SubdepartmentImporter(FromCOREPOSAPI, corepos_importing.model.SubdepartmentImporter):
    """
    Importer for subdepartment data from CORE POS API.
    """
    key = 'corepos_number'
    supported_fields = [
        'corepos_number',
        'number',
        'name',
        'department_number',
    ]

    def get_host_objects(self):
        return self.api.get_subdepartments()

    def normalize_host_object(self, subdepartment):
        department_number = None
        if 'dept_ID' in subdepartment:
            department_number = int(subdepartment['dept_ID'])

        return {
            'corepos_number': int(subdepartment['subdept_no']),
            'number': int(subdepartment['subdept_no']),
            'name': subdepartment['subdept_name'],
            'department_number': department_number,
        }


class VendorImporter(FromCOREPOSAPI, corepos_importing.model.VendorImporter):
    """
    Importer for vendor data from CORE POS API.
    """
    key = 'corepos_id'
    supported_fields = [
        'corepos_id',
        'id',
        'name',
        'abbreviation',
        'special_discount',
        'phone_number',
        'fax_number',
        'email_address',
    ]

    def get_host_objects(self):
        return self.api.get_vendors()

    def normalize_host_object(self, vendor):
        return {
            'corepos_id': int(vendor['vendorID']),
            'id': str(vendor['vendorID']),
            'name': vendor['vendorName'],
            'abbreviation': vendor.get('vendorAbbreviation') or None,
            'special_discount': decimal.Decimal(vendor['discountRate']),
            'phone_number': vendor.get('phone') or None,
            'fax_number': vendor.get('fax') or None,
            'email_address': vendor.get('email') or None,
        }


class ProductImporter(FromCOREPOSAPI, corepos_importing.model.ProductImporter):
    """
    Importer for product data from CORE POS API.
    """
    key = 'uuid'
    supported_fields = [
        'uuid',
        'corepos_id',
        'item_id',
        'upc',
        'brand_name',
        'description',
        'unit_size',
        'unit_of_measure',
        'uom_abbreviation',
        'size',
        'weighed',
        'department_number',
        'subdepartment_number',
        'regular_price_price',
        'regular_price_multiple',
        'regular_price_type',
        'food_stampable',
        # 'tax1',
        # 'tax2',
        'case_size',
    ]

    def setup(self):
        super(ProductImporter, self).setup()
        model = self.model

        query = self.session.query(model.Product)\
                            .join(model.CoreProduct)\
                            .filter(model.CoreProduct.corepos_id != None)\
                            .options(orm.joinedload(model.Product._corepos))
        self.core_existing = self.cache_model(model.Product, key='corepos_id',
                                              query=query)

        self.vendor_items_by_upc = {}

        def cache(item, i):
            self.vendor_items_by_upc.setdefault(item['upc'], []).append(item)

        self.progress_loop(cache, self.api.get_vendor_items(),
                           message="Caching CORE Vendor Items")

    def get_host_objects(self):
        return self.api.get_products()

    def identify_product(self, corepos_product):
        model = self.config.get_model()
        corepos_id = int(corepos_product['id'])

        if hasattr(self, 'core_existing'):
            product = self.core_existing.get(corepos_id)
            if product:
                return product

        else:
            try:
                return self.session.query(model.Product)\
                                   .join(model.CoreProduct)\
                                   .filter(model.CoreProduct.corepos_id == corepos_id)\
                                   .one()
            except orm.exc.NoResultFound:
                pass

        # at this point we'll search by `Product.item_id` instead
        return self.session.query(model.Product)\
                           .outerjoin(model.CoreProduct)\
                           .filter(model.CoreProduct.corepos_id == None)\
                           .filter(model.Product.item_id == corepos_product['upc'])\
                           .first()

    def identify_product_uuid(self, corepos_product):
        product = self.identify_product(corepos_product)
        if product:
            return product.uuid
        return get_uuid()

    def get_vendor_items(self, api_product):
        if hasattr(self, 'vendor_items_by_upc'):
            return self.vendor_items_by_upc.get(api_product['upc'])

        return self.api.get_vendor_items(upc=api_product['upc'])

    def normalize_host_object(self, product):
        try:
            upc = GPC(product['upc'], calc_check_digit='upc')
        except (TypeError, ValueError):
            log.debug("CORE POS product has invalid UPC: %s", product['upc'])
            if len(self.key) == 1 and self.key[0] == 'upc':
                return
            upc = None

        department_number = None
        if 'department' in product:
            department_number = int(product['department']) or None

        subdepartment_number = None
        if 'subdept' in product:
            subdepartment_number = int(product['subdept']) or None

        price = None
        if product.get('normal_price') is not None:
            price = decimal.Decimal(product['normal_price'])

        data = {
            'uuid': self.identify_product_uuid(product),
            'corepos_id': int(product['id']),
            'item_id': product['upc'],
            'upc': upc,
            'brand_name': product.get('brand') or None,
            'description': product.get('description') or '',

            'department_number': department_number,
            'subdepartment_number': subdepartment_number,

            'weighed': product['scale'] == '1',
            'food_stampable': product['foodstamp'] == '1',
            # 'tax1': product['tax'] == '1', # TODO: is this right?
            # 'tax2': product['tax'] == '2', # TODO: is this right?

            'regular_price_price': price,
            'regular_price_multiple': 1 if price is not None else None,
            'regular_price_type': self.enum.PRICE_TYPE_REGULAR if price is not None else None,
        }

        if self.fields_active(self.size_fields):
            size_info = self.normalize_size_info(product)
            data.update({
                'size': size_info['size'],
                'unit_size': size_info['unit_size'],
                'unit_of_measure': size_info['uom_code'],
                'uom_abbreviation': size_info['uom_abbrev'],
            })

        if 'case_size' in self.fields:
            case_size = None
            items = self.get_vendor_items(product)
            if items:
                # here we sort by `modified DESC` to get the "deterministic
                # pseudo-default" vendorItems record

                def sortkey(item):
                    modified = item.get('modified')
                    if not modified:
                        return datetime.datetime(1900, 1, 1)
                    dt = datetime.datetime.strptime(modified, '%Y-%m-%d %H:%M:%S')
                    return dt

                items = sorted(items, key=sortkey, reverse=True)
                item = items[0]
                case_size = decimal.Decimal(item['units'])

            data['case_size'] = case_size

        return data


class ProductMovementImporter(FromCOREPOSAPI, corepos_importing.model.ProductImporter):
    """
    Importer for product movement data from CORE POS API.
    """
    key = 'corepos_id'
    supported_fields = [
        'corepos_id',
        'last_sold',
    ]
    allow_create = False
    allow_delete = False

    def get_host_objects(self):
        return self.api.get_products()

    def normalize_host_object(self, product):

        last_sold = None
        if 'last_sold' in product:
            last_sold = datetime.datetime.strptime(product['last_sold'], '%Y-%m-%d %H:%M:%S')
            last_sold = localtime(self.config, last_sold)
            last_sold = make_utc(last_sold)

        return {
            'corepos_id': int(product['id']),
            'last_sold': last_sold,
        }


class ProductCostImporter(FromCOREPOSAPI, corepos_importing.model.ProductCostImporter):
    """
    Importer for product cost data from CORE POS API.
    """
    key = ('vendor_uuid', 'code')
    supported_fields = [
        'corepos_id',
        'product_uuid',
        'vendor_uuid',
        'code',
        'case_size',
        'case_cost',
        'unit_cost',
        'preferred',
    ]

    def setup(self):
        super(ProductCostImporter, self).setup()
        model = self.config.get_model()

        query = self.session.query(model.Vendor)\
                            .join(model.CoreVendor)\
                            .filter(model.CoreVendor.corepos_id != None)
        self.vendors = self.cache_model(model.Vendor, query=query,
                                        key='corepos_id')

        self.corepos_products = {}

        def cache(product, i):
            self.corepos_products[product['upc']] = product

        self.progress_loop(cache, self.api.get_products(),
                           message="Caching Products from CORE-POS API")

        self.products_by_item_id = self.cache_model(model.Product,
                                                    key='item_id')

    def get_host_objects(self):
        return self.api.get_vendor_items()

    def get_vendor(self, item):
        corepos_id = int(item['vendorID'])

        if hasattr(self, 'vendors'):
            return self.vendors.get(corepos_id)

        model = self.config.get_model()
        try:
            return self.session.query(model.Vendor)\
                               .join(model.CoreVendor)\
                               .filter(model.CoreVendor.corepos_id == corepos_id)\
                               .one()
        except orm.exc.NoResultFound:
            pass

    def get_corepos_product(self, item):
        if hasattr(self, 'corepos_products'):
            return self.corepos_products.get(item['upc'])

        return self.api.get_product(item['upc'])

    def get_product(self, item):
        item_id = item['upc']

        if hasattr(self, 'products_by_item_id'):
            return self.products_by_item_id.get(item_id)

        model = self.model
        try:
            return self.session.query(model.Product)\
                               .filter(model.Product.item_id == item_id)\
                               .one()
        except orm.exc.NoResultFound:
            pass

    def normalize_host_object(self, item):
        vendor = self.get_vendor(item)
        if not vendor:
            log.warning("CORE POS vendor not found for item: %s", item)
            return

        product = self.get_product(item)
        if not product:
            log.warning("product not found for CORE vendor item: %s", item)
            return

        core_product = self.get_corepos_product(item)
        # if not product:
        #     log.warning("CORE POS product not found for item: %s", item)
        #     return

        preferred = False
        if core_product and core_product['default_vendor_id'] == item['vendorID']:
            preferred = True

        case_size = decimal.Decimal(item['units'])
        unit_cost = decimal.Decimal(item['cost'])

        return {
            'corepos_id': int(item['vendorItemID']),
            'product_uuid': product.uuid,
            'vendor_uuid': vendor.uuid,
            'code': (item['sku'] or '').strip() or None,
            'case_size': case_size,
            'case_cost': case_size * unit_cost,
            'unit_cost': unit_cost,
            'preferred': preferred,
        }


class MemberImporter(FromCOREPOSAPI, corepos_importing.model.MemberImporter):
    """
    Importer for member data from CORE POS API.
    """
    key = 'corepos_account_id'
    supported_fields = [
        'corepos_account_id',
        'number',
        'id',
        'customer_uuid',
        'person_uuid',
        'joined',
        'withdrew',
    ]

    # TODO: should make this configurable
    member_status_codes = [
        'PC',
        'TERM',
    ]

    # TODO: should make this configurable
    non_member_status_codes = [
        'REG',
        'INACT',
    ]

    def setup(self):
        super(MemberImporter, self).setup()
        model = self.config.get_model()
        self.customers = self.cache_model(model.Customer, key='number')

    def get_host_objects(self):
        return self.get_core_members()

    def get_customer(self, number):
        if hasattr(self, 'customers'):
            return self.customers.get(number)

        model = self.config.get_model()
        try:
            return self.session.query(model.Customer)\
                               .filter(model.Customer.number == number)\
                               .one()
        except orm.exc.NoResultFound:
            pass

    def normalize_host_object(self, member):
        customer = self.get_customer(member['cardNo'])
        if not customer:
            log.warning("Rattail customer not found for cardNo %s: %s",
                        member['cardNo'], member)
            return

        person = customer.first_person()
        if not person:
            log.warning("Rattail person not found for cardNo %s: %s",
                        member['cardNo'], member)
            return

        if member['memberStatus'] in self.non_member_status_codes:
            log.debug("skipping non-member %s with status '%s': %s",
                      member['memberStatus'], member['cardNo'], member)
            return
        if member['memberStatus'] not in self.member_status_codes:
            # note that we will still import this one! we don't skip it
            log.warning("unexpected status '%s' for member %s: %s",
                        member['memberStatus'], member['cardNo'], member)

        joined = None
        if member['startDate'] and member['startDate'] != '0000-00-00 00:00:00':
            joined = datetime.datetime.strptime(member['startDate'],
                                                '%Y-%m-%d %H:%M:%S')
            joined = joined.date()

        withdrew = None
        if member['endDate'] and member['endDate'] != '0000-00-00 00:00:00':
            withdrew = datetime.datetime.strptime(member['endDate'],
                                                  '%Y-%m-%d %H:%M:%S')
            withdrew = withdrew.date()

        return {
            'corepos_account_id': int(member['customerAccountID']),
            'number': int(member['cardNo']),
            'id': str(member['customerAccountID']),
            'customer_uuid': customer.uuid,
            'person_uuid': person.uuid,
            'joined': joined,
            'withdrew': withdrew,
        }



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CSV -> CORE data import
"""

from corepos.db.office_op import model as corepos, Session as CoreSession

from rattail.importing.handlers import FromFileHandler
from rattail.importing.csv import FromCSVToSQLAlchemyMixin
from rattail_corepos.corepos.importing.db.model import ToCore
from rattail_corepos.corepos.importing.db.corepos import ToCoreHandler


class FromCSVToCore(FromCSVToSQLAlchemyMixin, FromFileHandler, ToCoreHandler):
    """
    Handler for CSV -> CORE data import
    """
    host_title = "CSV"
    ToParent = ToCore

    @property
    def local_title(self):
        return "CORE ({})".format(self.dbkey)

    def get_model(self):
        return corepos

    def make_session(self):
        return CoreSession(bind=self.config.corepos_engines[self.dbkey])



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE Office commands
"""

import sys

from rattail import commands
from rattail_corepos import __version__
from rattail.util import load_object


def main(*args):
    """
    Entry point for 'core-office' commands
    """
    if args:
        args = list(args)
    else:
        args = sys.argv[1:]

    cmd = Command()
    cmd.run(*args)


class Command(commands.Command):
    """
    Primary command for CORE Office
    """
    name = 'core-office'
    version = __version__
    description = "core-office -- command line interface for CORE Office"
    long_description = ""


class ExportLaneOp(commands.ImportSubcommand):
    """
    Export "op" data from CORE Office to CORE Lane
    """
    name = 'export-lane-op'
    description = __doc__.strip()
    default_handler_spec = 'rattail_corepos.corepos.lane.importing.op.office:FromCoreOfficeToCoreLane'
    default_dbkey = 'default'

    def get_handler_factory(self, **kwargs):
        if self.config:
            spec = self.config.get('corepos.lane.importing', 'office.handler',
                                   default=self.default_handler_spec)
        else:
            # just use default, for sake of cmd line help
            spec = self.default_handler_spec
        return load_object(spec)

    def add_parser_args(self, parser):
        super(ExportLaneOp, self).add_parser_args(parser)
        parser.add_argument('--dbkey', metavar='KEY', default=self.default_dbkey,
                            help="Config key for database engine to be used as the "
                            "\"target\" CORE Lane DB, i.e. where data will be "
                            " exported.  This key must be defined in the "
                            " [rattail_corepos.db.lane_op] section of your "
                            "config file.")

    def get_handler_kwargs(self, **kwargs):
        if 'args' in kwargs:
            kwargs['dbkey'] = kwargs['args'].dbkey
        return kwargs



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Rattail -> CORE-POS data export
"""

import logging

from sqlalchemy import orm

from rattail import importing
from rattail.db import model
from rattail.db.util import short_session
from rattail.util import OrderedDict, pretty_quantity
from rattail_corepos.corepos import importing as corepos_importing
from rattail_corepos.corepos.util import get_max_existing_vendor_id


log = logging.getLogger(__name__)


class FromRattailToCore(importing.FromRattailHandler):
    """
    Rattail -> CORE-POS export handler
    """
    host_title = "Rattail"
    local_title = "CORE-POS"
    direction = 'export'

    def get_importers(self):
        importers = OrderedDict()
        importers['Member'] = MemberImporter
        importers['Department'] = DepartmentImporter
        importers['Subdepartment'] = SubdepartmentImporter
        importers['Vendor'] = VendorImporter
        importers['Product'] = ProductImporter
        return importers


class FromRattail(importing.FromSQLAlchemy):
    """
    Base class for Rattail -> CORE-POS exporters.
    """


class MemberImporter(FromRattail, corepos_importing.model.MemberImporter):
    """
    Member data exporter
    """
    host_model_class = model.Customer
    key = 'cardNo'
    supported_fields = [
        'cardNo',
        'customerAccountID',
        'customers',
        'addressFirstLine',
        'addressSecondLine',
        'city',
        'state',
        'zip',
        'startDate',
        'endDate',
    ]
    supported_customer_fields = [
        'customerID',
        'firstName',
        'lastName',
        'accountHolder',
        'phone',
        'altPhone',
        'email',
    ]

    def query(self):
        query = super(MemberImporter, self).query()
        query = query.options(orm.joinedload(model.Customer.addresses))\
                .options(orm.joinedload(model.Customer._people)\
                         .joinedload(model.CustomerPerson.person)\
                         .joinedload(model.Person.phones))\
                .options(orm.joinedload(model.Customer._people)\
                         .joinedload(model.CustomerPerson.person)\
                         .joinedload(model.Person.emails))
        return query

    def normalize_host_object(self, customer):

        address = customer.addresses[0] if customer.addresses else None

        people = []
        for i, person in enumerate(customer.people, 1):
            phones = person.phones
            phone1 = phones[0] if phones else None
            phone2 = phones[1] if len(phones) > 1 else None
            email = person.emails[0] if person.emails else None
            people.append({
                'customerID': str(person.corepos_customer_id),
                'firstName': person.first_name,
                'lastName': person.last_name,
                'accountHolder': i == 1,
                'phone': phone1.number if phone1 else '',
                'altPhone': phone2.number if phone2 else '',
                'email': email.address if email else '',
            })

        member = customer.only_member(require=False)
        if member:
            if member.joined:
                start_date = member.joined.strftime('%Y-%m-%d 00:00:00')
            else:
                start_date = self.empty_date_value
            if member.withdrew:
                end_date = member.withdrew.strftime('%Y-%m-%d 00:00:00')
            else:
                end_date = self.empty_date_value
        else:
            start_date = '__omit__'
            end_date = '__omit__'

        return {
            'cardNo': customer.number,
            'customerAccountID': customer.id,
            'addressFirstLine': address.street if address else '',
            'addressSecondLine': address.street2 if address else '',
            'city': address.city if address else '',
            'state': address.state if address else '',
            'zip': address.zipcode if address else '',
            'startDate': start_date,
            'endDate': end_date,
            'customers': people,
        }


class DepartmentImporter(FromRattail, corepos_importing.model.DepartmentImporter):
    """
    Department data exporter
    """
    host_model_class = model.Department
    key = 'dept_no'
    supported_fields = [
        'dept_no',
        'dept_name',
    ]

    def normalize_host_object(self, department):
        return {
            'dept_no': str(department.number),
            'dept_name': department.name,
        }


class SubdepartmentImporter(FromRattail, corepos_importing.model.SubdepartmentImporter):
    """
    Subdepartment data exporter
    """
    host_model_class = model.Subdepartment
    key = 'subdept_no'
    supported_fields = [
        'subdept_no',
        'subdept_name',
        'dept_ID',
    ]

    def normalize_host_object(self, subdepartment):
        department = subdepartment.department
        return {
            'subdept_no': str(subdepartment.number),
            'subdept_name': subdepartment.name,
            'dept_ID': str(department.number) if department else None,
        }


class VendorImporter(FromRattail, corepos_importing.model.VendorImporter):
    """
    Vendor data exporter
    """
    host_model_class = model.Vendor
    key = 'vendorID'
    supported_fields = [
        'vendorID',
        'vendorName',
        'vendorAbbreviation',
        'discountRate',
        'phone',
        'fax',
        'email',
    ]

    def setup(self):
        super(VendorImporter, self).setup()

        # self.max_existing_vendor_id = self.get_max_existing_vendor_id()
        self.max_existing_vendor_id = get_max_existing_vendor_id()
        self.last_vendor_id = self.max_existing_vendor_id

    def get_next_vendor_id(self):
        if hasattr(self, 'last_vendor_id'):
            self.last_vendor_id += 1
            return self.last_vendor_id

        last_vendor_id = get_max_existing_vendor_id()
        return last_vendor_id + 1

    def normalize_host_object(self, vendor):
        vendor_id = vendor.corepos_id
        if not vendor_id:
            vendor_id = self.get_next_vendor_id()

        data = {
            'vendorID': str(vendor_id),
            'vendorName': vendor.name,
            'vendorAbbreviation': vendor.abbreviation or '',
            'discountRate': float(vendor.special_discount or 0),
        }

        if 'phone' in self.fields:
            phones = [phone for phone in vendor.phones
                      if phone.type == 'Voice']
            data['phone'] = phones[0].number if phones else ''

        if 'fax' in self.fields:
            phones = [phone for phone in vendor.phones
                      if phone.type == 'Fax']
            data['fax'] = phones[0].number if phones else ''

        if 'email' in self.fields:
            email = vendor.email
            data['email'] = email.address if email else ''

        # also embed original Rattail vendor object, if we'll be needing to
        # update it later with a new CORE ID
        if not vendor.corepos_id:
            data['_rattail_vendor'] = vendor

        return data

    def create_object(self, key, data):

        # grab vendor object we (maybe) stashed when normalizing
        rattail_vendor = data.pop('_rattail_vendor', None)

        # do normal create logic
        vendor = super(VendorImporter, self).create_object(key, data)
        if vendor:

            # maybe set the CORE ID for vendor in Rattail
            if rattail_vendor:
                rattail_vendor.corepos_id = int(vendor['vendorID'])

            return vendor


class ProductImporter(FromRattail, corepos_importing.model.ProductImporter):
    """
    Product data exporter
    """
    host_model_class = model.Product
    key = 'upc'
    supported_fields = [
        'upc',
        'brand',
        'description',
        'size',
        'unitofmeasure',
        'department',
        'normal_price',
        'foodstamp',
        'scale',
    ]

    def normalize_host_object(self, product):
        upc = product.item_id
        if not upc and product.upc:
            upc = str(product.upc)[:-1]
        if not upc:
            log.warning("skipping product %s with unknown upc: %s",
                        product.uuid, product)
            return

        return {
            '_product': product,
            'upc': upc,
            'brand': product.brand.name if product.brand else '',
            'description': product.description or '',
            'size': pretty_quantity(product.unit_size),
            'unitofmeasure': product.uom_abbreviation,
            'department': str(product.department.number) if product.department else None,
            'normal_price': '{:0.2f}'.format(product.regular_price.price) if product.regular_price else None,
            'foodstamp': '1' if product.food_stampable else '0',
            'scale': '1' if product.weighed else '0',
        }

    def create_object(self, key, data):

        # must be sure not to pass the original Product instance, or else the
        # API call will try to serialize and submit it
        product = data.pop('_product')

        corepos_product = super(ProductImporter, self).create_object(key, data)
        if corepos_product:

            # update our Rattail Product with the CORE ID
            if not self.dry_run:
                product.corepos_id = int(corepos_product['id'])
                return corepos_product

    def update_object(self, corepos_product, data, local_data=None):

        # must be sure not to pass the original Product instance, or else the
        # API call will try to serialize and submit it
        product = data.pop('_product', None)

        corepos_product = super(ProductImporter, self).update_object(corepos_product, data, local_data)
        return corepos_product



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
DataSync for Rattail DB
"""

from sqlalchemy.orm.exc import NoResultFound

from corepos.db.office_op import Session as CoreSession, model as corepos

from rattail.datasync import NewDataSyncImportConsumer
from rattail_corepos.corepos.api import make_corepos_api


class FromCOREAPIToRattail(NewDataSyncImportConsumer):
    """
    Consumer for CORE POS (API) -> Rattail datasync
    """
    handler_spec = 'rattail_corepos.importing.corepos.api:FromCOREPOSToRattail'
    model_map = {
        'VendorItem': 'ProductCost',
    }

    def setup(self):
        super(FromCOREAPIToRattail, self).setup()
        self.establish_api()

    def establish_api(self):
        self.api = make_corepos_api(self.config)

    def process_changes(self, session, changes):
        if self.runas_username:
            session.set_continuum_user(self.runas_username)

        # update all importers with current Rattail session
        for importer in self.importers.values():
            importer.session = session
            # also establish the API client for each!
            importer.establish_api()

        # sync all Customer-related changes
        types = [
            'Member',
        ]
        for change in [c for c in changes if c.payload_type in types]:
            if change.deletion:
                # normal logic works fine for this (maybe?)
                self.invoke_importer(session, change)
            else:
                # import member data from API, into various Rattail tables
                member = self.get_host_object(session, change)
                self.process_change(session, self.importers['Customer'],
                                    host_object=member)
                people = self.importers['Person'].get_person_objects_for_member(member)
                for person in people:
                    self.process_change(session, self.importers['Person'],
                                        host_object=person)
                self.process_change(session, self.importers['Member'],
                                    host_object=member)

        # sync all "product meta" changes
        types = [
            'Department',
            'Subdepartment',
            'Vendor',
        ]
        for change in [c for c in changes if c.payload_type in types]:
            # normal logic works fine here
            self.invoke_importer(session, change)

        # sync all Product-related changes
        types = [
            'Product',
            'VendorItem',
        ]
        for change in [c for c in changes if c.payload_type in types]:

            # process deletion per usual logic
            if change.deletion:
                self.invoke_importer(session, change)
                if change.payload_type == 'Product':
                    continue # stop here if we just deleted a Product

            # try to process change for Product
            product = self.get_product(change)
            if product:
                self.process_change(session, self.importers['Product'],
                                    host_object=product)

        # sync all ProductCost-related changes
        types = [
            'VendorItem',
        ]
        for change in [c for c in changes if c.payload_type in types]:
            # normal logic works fine here
            self.invoke_importer(session, change)

    def get_host_object(self, session, change):
        if change.payload_type == 'Member':
            return self.api.get_member(change.payload_key)
        if change.payload_type == 'Department':
            return self.api.get_department(change.payload_key)
        if change.payload_type == 'Subdepartment':
            return self.api.get_subdepartment(change.payload_key)
        if change.payload_type == 'Vendor':
            return self.api.get_vendor(change.payload_key)
        if change.payload_type == 'Product':
            return self.api.get_product(change.payload_key)
        if change.payload_type == 'VendorItem':
            fields = change.payload_key.split('|')
            if len(fields) == 2:
                sku, vendorID = fields
                if vendorID.isdigit():
                    return self.api.get_vendor_item(sku, int(vendorID))

    def get_product(self, change):
        if change.payload_type == 'Product':
            return self.api.get_product(change.payload_key)
        if change.payload_type == 'VendorItem':
            fields = change.payload_key.split('|')
            if len(fields) == 2:
                sku, vendorID = fields
                vendor_item = self.api.get_vendor_item(sku, vendorID)
                if vendor_item:
                    return self.api.get_product(vendor_item['upc'])


class FromCOREPOSToRattailBase(NewDataSyncImportConsumer):
    """
    Base class for CORE POS -> Rattail data sync consumers.
    """
    handler_spec = 'rattail_corepos.importing.corepos.db:FromCOREPOSToRattail'

    def begin_transaction(self):
        self.corepos_session = CoreSession()

    def rollback_transaction(self):
        self.corepos_session.rollback()
        self.corepos_session.close()

    def commit_transaction(self):
        # always rollback here, we don't want any accidents in CORE POS
        self.corepos_session.rollback()
        self.corepos_session.close()


class FromCOREPOSToRattailProducts(FromCOREPOSToRattailBase):
    """
    Handles CORE POS -> Rattail sync for product data.
    """

    def get_host_object(self, session, change):

        if change.payload_type == 'Product':
            try:
                return self.corepos_session.query(corepos.Product)\
                                           .filter(corepos.Product.upc == change.payload_key)\
                                           .one()
            except NoResultFound:
                pass

        else:
            # try to fetch CORE POS object via typical method
            Model = getattr(corepos, change.payload_type)
            return self.corepos_session.query(Model)\
                                       .get(int(change.payload_key))
